]> Shamusworld >> Repos - rmac/blobdiff - direct.c
EOY minor update.
[rmac] / direct.c
index 1864b86e266142964f630166ca37390a48a40d39..b3f51f3c3b73356c638a0dc9b087cc20f67ab58d 100644 (file)
--- a/direct.c
+++ b/direct.c
@@ -1,7 +1,7 @@
 //
 // RMAC - Reboot's Macro Assembler for all Atari computers
 // DIRECT.C - Directive Handling
-// Copyright (C) 199x Landon Dyer, 2011-2019 Reboot and Friends
+// Copyright (C) 199x Landon Dyer, 2011-2020 Reboot and Friends
 // RMAC derived from MADMAC v1.07 Written by Landon Dyer, 1986
 // Source utilised with the kind permission of Landon Dyer
 //
@@ -51,6 +51,7 @@ int d_abs(void);
 int d_comm(void);
 int d_dc(WORD);
 int d_ds(WORD);
+int d_dsm(WORD);
 int d_dcb(WORD);
 int d_globl(void);
 int d_gpu(void);
@@ -153,6 +154,7 @@ int (*dirtab[])() = {
        d_nofpu,                        // 65 nofpu
        d_opt,                          // 66 .opt
        d_objproc,                      // 67 .objproc
+       d_dsm,                          // 68 .dsm
 };
 
 
@@ -1074,11 +1076,11 @@ int d_ds(WORD siz)
                return 0;
 
        // Check to see if the value being passed in is negative (who the hell does
-       // that?--nobody does; it's the code gremlins, or rum, that does it)
-       // N.B.: Since 'eval' is of type uint32_t, if it goes negative, it will have
-       //       its high bit set.
-       if (eval & 0x80000000)
-               return error("negative sizes not allowed");
+       // that?--nobody does; it's the code gremlins, or rum, what does it)
+       // N.B.: Since 'eval' is of type uint64_t, if it goes negative, it will
+       //       have its high bit set.
+       if (eval & 0x8000000000000000)
+               return error("negative sizes not allowed in DS");
 
        // In non-TDB section (BSS, ABS and M6502) just advance the location
        // counter appropriately. In TDB sections, deposit (possibly large) chunks
@@ -1094,18 +1096,86 @@ int d_ds(WORD siz)
 
                just_bss = 1;                                   // No data deposited (8-bit CPU mode)
        }
+       else if (cursect & M56KPXYL)
+       {
+               // Change segment instead of marking blanks.
+               // Only mark segments we actually wrote something
+               if (chptr != dsp_currentorg->start && dsp_written_data_in_current_org)
+               {
+                       dsp_currentorg->end = chptr;
+                       dsp_currentorg++;
+                       dsp_currentorg->memtype = dsp_currentorg[-1].memtype;
+               }
+
+               listvalue((uint32_t)eval);
+               sloc += (uint32_t)eval;
+
+               // And now let's create a new segment
+               dsp_currentorg->start = chptr;
+               dsp_currentorg->chunk = scode;  // Mark down which chunk this org starts from (will be needed when outputting)
+               sect[cursect].orgaddr = sloc;
+               dsp_currentorg->orgadr = sloc;
+               dsp_written_data_in_current_org = 0;
+
+               just_bss = 1;                                   // No data deposited
+       }
        else
        {
-               dep_block(eval, siz, 0, (WORD)(DEFINED | ABS), NULL);
+               dep_block(eval, siz, 0, (DEFINED | ABS), NULL);
        }
 
        ErrorIfNotAtEOL();
-       return 0;
+       return OK;
+}
+
+
+//
+// dsm[.siz] expression
+// Define modulo storage
+// Quoting the Motorola assembler manual:
+// "The DSM directive reserves a block of memory the length of which in words is equal to
+// the value of <expression>.If the runtime location counter is not zero, this directive first
+// advances the runtime location counter to a base address that is a multiple of 2k, where
+// 2k >= <expression>."
+// The kicker of course is written a few sentences after:
+// "<label>, if present, will be assigned the value of the runtime location counter after a valid
+// base address has been established."
+//
+int d_dsm(WORD siz)
+{
+       TOKEN * tok_current = tok;  // Keep track of where tok was when we entered this procedure
+       uint64_t eval;
+
+       if (abs_expr(&eval) != OK)
+               return 0;
+
+       // Round up to the next highest power of 2
+       // Nicked from https://graphics.stanford.edu/~seander/bithacks.html#RoundUpPowerOf2
+       eval--;
+       eval |= eval >> 1;
+       eval |= eval >> 2;
+       eval |= eval >> 4;
+       eval |= eval >> 8;
+       eval |= eval >> 16;
+
+       int units_to_skip;
+       units_to_skip = eval + 1 - sloc;
+       sloc += units_to_skip;          // Bump up sloc - TODO: check if this goes over the RAM limits?
+
+       // If a label has been defined in the same line as dsm, its value also needs to be adjusted
+       if (label_defined)
+       {
+               SYM * label = lookup(label_defined, LABEL, 0);
+               label->svalue += units_to_skip;
+       }
+
+       tok = tok_current;              // Rewind tok back to where it was
+       return d_ds(siz);               // And let d_ds take over from here
 }
 
 
 //
-// dc.b, dc.w / dc, dc.l, dc.i, dc.q, dc.d
+// dc.b, dc.w / dc, dc.l, dc.i, dc.q, dc.d, dc.s, dc.x
 //
 int d_dc(WORD siz)
 {
@@ -1289,6 +1359,7 @@ l_parse_loop:
                        }
 
                        break;
+
                case SIZW:
                case SIZN:
                        if (!defined)
@@ -1312,6 +1383,7 @@ l_parse_loop:
                        }
 
                        break;
+
                case SIZL:
                        // Shamus: Why can't we do longs in 6502 mode?
                        if (m6502)
@@ -1319,11 +1391,7 @@ l_parse_loop:
 
                        if (!defined)
                        {
-                               if (movei)
-                                       AddFixup(FU_LONG | FU_MOVEI, sloc, exprbuf);
-                               else
-                                       AddFixup(FU_LONG, sloc, exprbuf);
-
+                               AddFixup(FU_LONG | (movei ? FU_MOVEI : 0), sloc, exprbuf);
                                D_long(0);
                        }
                        else
@@ -1338,34 +1406,33 @@ l_parse_loop:
                        }
 
                        break;
+
                case SIZQ:
                        // 64-bit size
                        if (m6502)
                                return error(in_6502mode);
 
-                       // Shamus: We only handle DC.Q type stuff, will have to add fixups
-                       //         and stuff later (maybe... might not be needed...)
                        // DEFINITELY NEED FIXUPS HERE!
                        if (!defined)
                        {
                                AddFixup(FU_QUAD, sloc, exprbuf);
-                               D_quad(0LL);
-                       }
-                       else
-                       {
-                               D_quad(eval);
+                               eval = 0;
                        }
 
+                       D_quad(eval);
                        break;
+
                case SIZS:
                        // 32-bit float size
                        if (m6502)
                                return error(in_6502mode);
 
+/* Seems to me that if something is undefined here, then that should be an error.  Likewise for the D & X variants. */
                        if (!defined)
                        {
-                               AddFixup(FU_FLOATSING, sloc, exprbuf);
-                               D_long(0);
+//                             AddFixup(FU_FLOATSING, sloc, exprbuf);
+//                             D_long(0);
+                               return error("labels not allowed in floating point expressions");
                        }
                        else
                        {
@@ -1380,6 +1447,7 @@ l_parse_loop:
                        }
 
                        break;
+
                case SIZD:
                        // 64-bit double size
                        if (m6502)
@@ -1387,8 +1455,9 @@ l_parse_loop:
 
                        if (!defined)
                        {
-                               AddFixup(FU_FLOATDOUB, sloc, exprbuf);
-                               D_quad(0LL);
+//                             AddFixup(FU_FLOATDOUB, sloc, exprbuf);
+//                             D_quad(0LL);
+                               return error("labels not allowed in floating point expressions");
                        }
                        else
                        {
@@ -1403,6 +1472,7 @@ l_parse_loop:
                        }
 
                        break;
+
                case SIZX:
                        if (m6502)
                                return error(in_6502mode);
@@ -1412,8 +1482,9 @@ l_parse_loop:
 
                        if (!defined)
                        {
-                               AddFixup(FU_FLOATEXT, sloc, exprbuf);
-                               D_extend(extDbl);
+//                             AddFixup(FU_FLOATEXT, sloc, exprbuf);
+//                             D_extend(extDbl);
+                               return error("labels not allowed in floating point expressions");
                        }
                        else
                        {
@@ -1545,11 +1616,8 @@ int d_init(WORD def_siz)
 //
 int dep_block(uint32_t count, WORD siz, uint32_t eval, WORD eattr, TOKEN * exprbuf)
 {
-       WORD tdb;
-       WORD defined;
-
-       tdb = (WORD)(eattr & TDB);
-       defined = (WORD)(eattr & DEFINED);
+       WORD tdb = eattr & TDB;
+       WORD defined = eattr & DEFINED;
 
        while (count--)
        {